"""
TODO


"""

import numpy as np
from skimage.draw import polygon,polygon_perimeter
from skimage.measure import regionprops
from skimage import measure
from collections import Counter
import os
from natsort import natsorted
from epyseg.utils.loadlist import loadlist
from epyseg.ta.colors.colorgen import get_unique_random_color_int24
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()
import os

def get_TA_file(file_name_without_ext, TA_file):
    """
    Returns the path to the TA file by joining the file name without extension and the TA file.

    Args:
        file_name_without_ext (str): File name without extension.
        TA_file (str): TA file.

    Returns:
        str: Path to the TA file.

    """
    return os.path.join(file_name_without_ext, TA_file)


def get_n_consecutive_input_files_and_output_folders(images_to_analyze, start_idx, nb_of_images_to_get, incr=1):
    """
    Returns a list of consecutive input files and output folders.

    Args:
        images_to_analyze (list): List of images to analyze.
        start_idx (int): Starting index.
        nb_of_images_to_get (int): Number of consecutive images to get.
        incr (int, optional): Increment value. Defaults to 1.

    Returns:
        list: List of input files and output folders.

    """
    return get_n_input_files_and_output_folders(images_to_analyze, start_idx, 0, nb_of_images_to_get, incr=incr)


def get_n_input_files_and_output_folders(images_to_analyze, start_idx, start_range, end_range_inclusive, incr=1):
    """
    Returns a list of input files and output folders within a specified range.

    Args:
        images_to_analyze (list): List of images to analyze.
        start_idx (int): Starting index.
        start_range (int): Starting range.
        end_range_inclusive (int): Ending range (inclusive).
        incr (int, optional): Increment value. Defaults to 1.

    Returns:
        list: List of input files and output folders.

    """
    try:
        return images_to_analyze[start_idx + start_range: start_idx + end_range_inclusive: incr]
    except:
        output_data = []
        for i in range(start_range, end_range_inclusive + 1, incr):
            output_data.append(get_input_file_and_output_folder(images_to_analyze, start_idx + i))
        return output_data


def get_n_files_from_list(images_to_analyze, start_idx, start_range, end_range_inclusive, incr=1):
    """
    Returns a list of files from a given list within a specified range.

    Args:
        images_to_analyze (list): List of images to analyze.
        start_idx (int): Starting index.
        start_range (int): Starting range.
        end_range_inclusive (int): Ending range (inclusive).
        incr (int, optional): Increment value. Defaults to 1.

    Returns:
        list: List of files.

    Raises:
        Exception: Raises an exception if the list is empty.

    """
    if images_to_analyze is None or not images_to_analyze:
        raise Exception('Empty list, nothing to do...')
    try:
        if start_idx + start_range < 0 or start_idx + end_range_inclusive >= len(images_to_analyze):
            raise IndexError("List index out of range")
        return images_to_analyze[start_idx + start_range: start_idx + end_range_inclusive + 1: incr]
    except:
        output_data = []
        for i in range(start_range, end_range_inclusive + 1, incr):
            if start_idx + i < 0 or start_idx + i >= len(images_to_analyze):
                output_data.append(None)
                continue
            output_data.append(images_to_analyze[start_idx + i])
        return output_data


known_parsers = ['parent',
                'ext',
                'short',
                'full_no_ext',
                'TA',  # same as full no ext
                'short_no_ext',
                'full',
                'full_no_end_slash',
                'short_pop_all_exts',  # extreme version of short no ext where series of exts are removed, e.g. hiC_250522.pairs.txt.gz -−> hiC_250522
                ]


def replace_empty_string_by_none(input_string):
    """
    Replaces an empty string with None.

    Args:
        input_string (str): Input string.

    Returns:
        str: Input string with empty string replaced by None.

    """
    if not input_string:
        return None
    return input_string


def smart_name_appender(name, condition_boolean, text_to_append_if_condition_is_true=None,
                        text_to_append_if_condition_is_false=None):
    """
    Appends a string to a name depending on a condition.

    Args:
        name (str): Name to append the string.
        condition_boolean (bool): Condition.
        text_to_append_if_condition_is_true (str, optional): Text to append if the condition is True. Defaults to None.
        text_to_append_if_condition_is_false (str, optional): Text to append if the condition is False. Defaults to None.

    Returns:
        str: Modified name.

    """
    if condition_boolean:
        if text_to_append_if_condition_is_true is not None:
            name += text_to_append_if_condition_is_true
    else:
        if text_to_append_if_condition_is_false is not None:
            name += text_to_append_if_condition_is_false
    return name


def smart_name_parser(full_file_path, ordered_output=known_parsers, appenders=None, replace_empty_by_none=False):
    """
    Parses the full file path and returns the specified output.

    Args:
        full_file_path (str): Full file path.
        ordered_output (str or list, optional): Ordered list of output formats. Defaults to known_parsers.
        appenders (str or list, optional): String or list of strings to append to the output. Defaults to None.
        replace_empty_by_none (bool, optional): Replace empty strings by None. Defaults to False.

    Returns:
        str or list: Parsed output.

    """
    if full_file_path is None:
        logger.error('Please specify an input file name')
        return
    if ordered_output is None:
        logger.error('Parsing unspecified, nothing to do')
        return

    fixed_path = full_file_path.replace('\\\\', '/').replace('\\', '/')
    filename0_without_path = os.path.basename(fixed_path)
    filename0_without_ext, ext = os.path.splitext(filename0_without_path)
    parent_dir_of_filename0 = os.path.dirname(fixed_path)

    single_output = False
    if isinstance(ordered_output, str):
        single_output = True
        ordered_output = [ordered_output]

    sorted_output = []
    for parser in ordered_output:
        if parser not in known_parsers:
            sorted_output.append(os.path.join(parent_dir_of_filename0, os.path.join(filename0_without_ext, parser)))
            continue
        if parser == 'parent':
            sorted_output.append(parent_dir_of_filename0)
            continue
        if parser == 'short_no_ext':
            sorted_output.append(filename0_without_ext)
            continue
        if parser == 'short_pop_all_exts':
            extracted = filename0_without_ext
            while '.' in extracted:
                extracted, _ = os.path.splitext(extracted)
            sorted_output.append(extracted)
        if parser == 'full_no_ext' or parser == 'TA':
            sorted_output.append(os.path.join(parent_dir_of_filename0, filename0_without_ext))
            continue
        if parser == 'short':
            sorted_output.append(filename0_without_path)
            continue
        if parser == 'full':
            sorted_output.append(full_file_path)
            continue
        if parser == 'ext':
            sorted_output.append(ext)
            continue
        if parser == 'full_no_end_slash':
            out = fixed_path
            if out.endswith('/'):
                out = out[:-1]
            sorted_output.append(out)

    if appenders is not None:
        if not isinstance(appenders, list):
            appenders = [appenders]
        sorted_output = [os.path.join(txt, *appenders) for txt in sorted_output]

    if replace_empty_by_none:
        sorted_output = [replace_empty_string_by_none(input_string) for input_string in sorted_output]

    if single_output:
        return sorted_output[0]
    return sorted_output


def get_input_file_and_output_folder(images_to_analyze, idx):
    """
    Returns the input file and output folder for a given index.

    Args:
        images_to_analyze (list): List of images to analyze.
        idx (int): Index of the image to retrieve.

    Returns:
        tuple: Input file path and output folder path.

    """
    file_path_0 = images_to_analyze[idx]
    filename0_without_ext = os.path.splitext(file_path_0)[0]
    return file_path_0, filename0_without_ext


def get_input_files_and_output_folders(images_to_analyze, start_idx, incr=1):
    """
    Returns the input files and output folders for consecutive indices.

    Args:
        images_to_analyze (list): List of images to analyze.
        start_idx (int): Starting index.
        incr (int, optional): Increment value. Defaults to 1.

    Returns:
        tuple: Input file paths and output folder paths.

    """
    file_path_0 = images_to_analyze[start_idx]
    file_path_1 = images_to_analyze[start_idx + incr]

    filename0_without_ext = os.path.splitext(file_path_0)[0]
    filename1_without_ext = os.path.splitext(file_path_1)[0]

    return file_path_0, file_path_1, filename0_without_ext, filename1_without_ext


def get_list_of_files(path):
    """
    Returns a list of files in the given path.

    Args:
        path (str or list): Path to the directory or a list of file paths.

    Returns:
        list: List of file paths.

    # Examples:
    #     >>> get_list_of_files('/path/to/files')
    #     ['/path/to/files/file1.txt', '/path/to/files/file2.txt', '/path/to/files/file3.txt']
    #
    #     >>> get_list_of_files(['/path/to/files/file1.txt', '/path/to/files/file2.txt'])
    #     ['/path/to/files/file1.txt', '/path/to/files/file2.txt']

    """
    if isinstance(path, list):
        return path

    if not path.lower().endswith('.txt') and not path.lower().endswith('.lst'):
        images_to_analyze = os.listdir(path)
        images_to_analyze = [os.path.join(path, f) for f in images_to_analyze if os.path.isfile(os.path.join(path, f))]
        images_to_analyze = natsorted(images_to_analyze)
    else:
        images_to_analyze = loadlist(path)

    return images_to_analyze


def first_image_tracking(mask_t0, labels_t0, regprps=None, assigned_ids=None, seed=None):
    """
    Performs tracking on the first image.

    Args:
        mask_t0 (ndarray): Binary mask of the first image.
        labels_t0 (ndarray): Labels of the first image.
        regprps (list, optional): List of region properties. Defaults to None.
        assigned_ids (list, optional): List of assigned IDs. Defaults to None.
        seed (int, optional): Random seed. Defaults to None.

    Returns:
        ndarray: Tracked image.

    """
    tracks_t0 = np.zeros_like(mask_t0, dtype=np.uint32)

    if regprps is None:
        regprps = regionprops(labels_t0)

    for iii, region in enumerate(regprps):
        # create a random color and store it
        if iii == 0:
            new_col = get_unique_random_color_int24(forbidden_colors=assigned_ids, seed_random=seed,
                                                    assign_new_col_to_forbidden=True)
        else:
            new_col = get_unique_random_color_int24(forbidden_colors=assigned_ids, assign_new_col_to_forbidden=True)

        tracks_t0[labels_t0 == iii + 1] = new_col

    tracks_t0[labels_t0 == 0] = 0xFFFFFF
    # save the image or return it
    return tracks_t0  # this is the image # need copy it


def deduplicate_duplicated_cells_using_random_ids(RGB24, labels=None, background=0xFFFFFF, assigned_ids=None):
    """
    Deduplicates duplicated cells using random IDs.

    Args:
        RGB24 (ndarray): RGB image.
        labels (ndarray, optional): Label image. Defaults to None.
        background (int, optional): Background color. Defaults to 0xFFFFFF.
        assigned_ids (list, optional): List of assigned IDs. Defaults to None.

    Returns:
        ndarray: Deduplicated RGB image.

    """
    if labels is None:
        labels = measure.label(RGB24, connectivity=1, background=background)

    cell_ids = []
    for region in regionprops(labels):
        cell_ids.append(RGB24[region.coords[0][0], region.coords[0][1]])

    cell_ids_n_count = Counter(cell_ids)
    duplicated_cells = {k: v for k, v in cell_ids_n_count.items() if v >= 2}

    duplicated_cells_keys = list(duplicated_cells.keys())
    for region in regionprops(labels):
        id = RGB24[region.coords[0][0], region.coords[0][1]]
        if id in duplicated_cells_keys:
            duplication_count = duplicated_cells[id]
            if duplication_count >= 2:
                duplicated_cells[id] = duplication_count - 1
                RGB24[labels == region.label] = get_unique_random_color_int24(forbidden_colors=assigned_ids,
                                                                              assign_new_col_to_forbidden=True)
            else:
                duplicated_cells_keys.remove(id)

    return RGB24


def assign_random_ID_to_missing_cells(RGB24img, labels, regprps=None, assigned_ids=None):
    """
    Assigns random IDs to missing cells in the image.

    Args:
        RGB24img (ndarray): RGB image.
        labels (ndarray): Label image.
        regprps (list, optional): List of region properties. Defaults to None.
        assigned_ids (list, optional): List of assigned IDs. Defaults to None.

    Returns:
        ndarray: RGB image with assigned random IDs.

    """
    if regprps is None:
        regprps = regionprops(labels)

    for iii, region in enumerate(regprps):
        color_of_first_pixel_of_potential_match = RGB24img[region.coords[0][0], region.coords[0][1]]

        if color_of_first_pixel_of_potential_match == 0:
            new_col = get_unique_random_color_int24(forbidden_colors=assigned_ids, assign_new_col_to_forbidden=True)
            RGB24img[labels == labels[region.coords[0][0], region.coords[0][1]]] = new_col

    return RGB24img


def get_lost_cells_between_first_and_second_set(first_set, second_set):
    """
    Returns the lost cells between the first and second set.

    Args:
        first_set (list or set): First set of cells.
        second_set (list or set): Second set of cells.

    Returns:
        set: Set of lost cells.

    Examples:
        >>> get_lost_cells_between_first_and_second_set([1, 2, 3, 4, 5], [2, 4, 6, 8])
        {1, 3, 5}

        >>> get_lost_cells_between_first_and_second_set({1, 2, 3, 4, 5}, {2, 4, 6, 8})
        {1, 3, 5}

    """
    set_t0 = first_set
    if not isinstance(set_t0, set):
        set_t0 = set(set_t0)
    set_t1 = second_set
    if not isinstance(set_t1, set):
        set_t1 = set(set_t1)
    return set_t0.difference(set_t1)


def get_common_cells(first_set, second_set):
    """
    Returns the common cells between the first and second set.

    Args:
        first_set (list or set): First set of cells.
        second_set (list or set): Second set of cells.

    Returns:
        set: Set of common cells.

    Examples:
        >>> get_common_cells([1, 2, 3, 4, 5], [2, 4, 6, 8])
        {2, 4}

        >>> get_common_cells({1, 2, 3, 4, 5}, {2, 4, 6, 8})
        {2, 4}

    """
    set_t0 = first_set
    if not isinstance(set_t0, set):
        set_t0 = set(set_t0)
    set_t1 = second_set
    if not isinstance(set_t1, set):
        set_t1 = set(set_t1)
    return set_t0.intersection(set_t1)


def get_cells_in_image(RGB24_img):
    """
    Returns the unique IDs of cells in the image.

    Args:
        RGB24_img (ndarray): RGB image.

    Returns:
        ndarray: Array of unique cell IDs.

    """
    unique_ids = np.unique(RGB24_img)
    return unique_ids


def get_cells_in_image_n_fisrt_pixel(RGB24_img):
    """
    Returns the unique IDs of cells in the image and their corresponding first pixel indices.

    Args:
        RGB24_img (ndarray): RGB image.

    Returns:
        ndarray: Array of unique cell IDs.
        ndarray: Array of indices of the first pixels of the cells.

    """
    u, indices = np.unique(RGB24_img, return_index=True)
    return u, indices


def get_cells_in_image_n_count(RGB24_img):
    """
    Returns the unique IDs of cells in the image and their corresponding counts.

    Args:
        RGB24_img (ndarray): RGB image.

    Returns:
        ndarray: Array of unique cell IDs.
        ndarray: Array of counts for each cell ID.

    """
    u, counts = np.unique(RGB24_img, return_counts=True)
    return u, counts


def plot_triangles(img, coords, tri, inner_color, perimeter_color=None):
    """
    Plot the triangles of a cell on an image.

    Args:
        img (ndarray): Image to plot on.
        coords (ndarray): Coordinates of the cell vertices.
        tri (Delaunay): Delaunay triangulation of the cell.
        inner_color (int): Inner color of the triangles.
        perimeter_color (int, optional): Perimeter color of the triangles. Defaults to None.

    """
    for simplex in tri.simplices:
        r = coords[simplex][:, 0]  # get the y coords of the simplex
        c = coords[simplex][:, 1]  # get the x coords of the simplex
        rr, cc = polygon(r, c)
        img[rr, cc] = inner_color
        if perimeter_color is not None:
            rr, cc = polygon_perimeter(r, c)
            img[rr, cc] = perimeter_color


if __name__ == '__main__':
    if True:
        tst_name = '/F/hiC_microC/haplotype_resolved_embryo/GSE121255_RAW/GSM3428927_2_4hrs-R2.dedup.pairs.txt.gz'

        print(smart_name_parser(tst_name,'short_pop_all_exts'))

        tst_name = '/F/hiC_microC/haplotype_resolved_embryo/GSE121255_RAW/GSM3428927_2_4hrs-R2'
        print(smart_name_parser(tst_name, 'short_pop_all_exts'))

        tst_name = '/F/hiC_microC/haplotype_resolved_embryo/GSE121255_RAW/GSM3428927_2_4hrs-R2.txt'
        print(smart_name_parser(tst_name, 'short_pop_all_exts'))

        import sys
        sys.exit(0)


    if True:
        print(smart_name_parser('Série 10kb X1 Spot196 mut2 TgX malex.tif.tif', 'tutu'))

        import sys
        sys.exit(0)

    if True:
        # these are the tests for the smart_name_parser

        print(smart_name_parser('/this/is/a/test/of/your/system.tif', 'parent', appenders=['toto', 'tutu',
                                                                                           'tata.tif']))  
        print(smart_name_parser('/this/is/a/test/of/your/system.tif', 'parent',
                                appenders='tata.tif'))  
        print(smart_name_parser('/this/is/a/test/of/your/system.tif', 'full_no_ext',
                                appenders='tata.tif'))  
        print(smart_name_parser('/this/is/a/test/of/your/system.tif', ['TA', 'parent'],
                                appenders='tata.tif'))  

        print('smart_name_parser', smart_name_parser(
            '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png'))  # ['/E/Sample_images/sample_images_PA/trash_test_mem/mini', '', 'focused_Series012', '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012', 'focused_Series012', '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012']
        print('smart_name_parser', smart_name_parser(
            '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012'))  # ['/E/Sample_images/sample_images_PA/trash_test_mem/mini', '', 'focused_Series012', '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012', 'focused_Series012', '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012']
        print('smart_name_parser', smart_name_parser(
            'focused_Series012.png'))  # smart_name_parser ['', '.png', 'focused_Series012.png', 'focused_Series012', 'focused_Series012', 'focused_Series012.png']
        print('smart_name_parser',
              smart_name_parser('D:\\Sample_images\\sample_images_PA\\trash_test_mem\\mini\\focused_Series012.png'))
        print('smart_name_parser', smart_name_parser(
            'D:\\Sample_images\\\\sample_images_PA/trash_test_mem\\mini\\focused_Series012.png'))  # hybrid widows linux names
        print('smart_name_parser',
              smart_name_parser('D:\\Sample_images\\\\sample_images_PA/trash_test_mem\\mini\\focused_Series012'))
        print('smart_name_parser', smart_name_parser(''))
        print('smart_name_parser', smart_name_parser('', replace_empty_by_none=True))
        print('smart_name_parser', smart_name_parser(None))
        print('smart_name_parser',
              smart_name_parser('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png',
                                ordered_output=[
                                    'tracked_cells_resized.png']))  # retruns a TA name --> it is really cool!!!
        tracked_cells_resized, TA_path = smart_name_parser('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png', ordered_output=['full_no_ext', 'tracked_cells_resized.png'])
        print('TADA', tracked_cells_resized, TA_path)
        tracked_cells_resized, TA_path = smart_name_parser('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png', ordered_output=['TA', 'tracked_cells_resized.png'])
        print('TADA2', tracked_cells_resized, TA_path)
        import sys
        sys.exit(0)

    if True:
        # test of getting n consecutive images
        # also see how to handle None --> TODO
        img_list = loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list.lst')
        print(img_list)

        desired_images = get_n_input_files_and_output_folders(img_list, 1, -1, 1) # get one image before and after

        print(desired_images)
        import sys
        sys.exit(0)


    # if True:
    #     random_vx_images = (np.random.randint(2, size=(1024, 1024)) * 255).astype('uint8')
    #     cell_labels = (np.random.randint(255, size=(1024, 1024))).astype('uint8')
    #     vertices = np.where(random_vx_images == 255)  # c'est
    #     count_same = 0
    #     count_different = 0
    #
    #     start_all = timer()
    #     for i in range(len(vertices[0])):
    #         y = vertices[0][i]
    #         x = vertices[1][i]
    #         # ids = neighbors8((y, x), cell_labels).tolist() # celui ci est faux
    #         ids2 = neighbors8_2((y, x), cell_labels)
    #
    #         # if  ids!=ids2:
    #         #     print(ids, ids2, ids==ids2)
    #
    #     print('time', timer() - start_all)
    #     # start_all = timer()
    #     # for i in range(len(vertices[0])):
    #     #     y = vertices[0][i]
    #     #     x = vertices[1][i]
    #     #     ids = neighbors8((y, x), cell_labels).tolist()  # celui ci est faux # --> keep my old way but fix its errors
    #     #     ids2 = neighbors8_2((y, x), cell_labels)
    #     #
    #     #     if ids != ids2:
    #     #         print(ids, ids2, ids == ids2, y, x)
    #     #         count_different+=1
    #     #     else:
    #     #         count_same+=1
    #     #
    #     # print(count_same, count_different)
    #     # print('time', timer() - start_all)
    #
    #     # this one is much faster but why ???
    #     start_all = timer()
    #     for i in range(len(vertices[0])):
    #         y = vertices[0][i]
    #         x = vertices[1][i]
    #         ids = neighbors8((y, x), cell_labels).tolist()  # celui ci est faux # --> keep my old way but fix its errors
    #         # ids2 = neighbors8_2((y, x), cell_labels)
    #
    #         # if  ids!=ids2:
    #         #     print(ids, ids2, ids==ids2)
    #
    #     print('time', timer() - start_all)



    cells_at_t0 = [0, 1, 2, 3, 4, 5, 6, 7, 128]
    cells_at_t1 = [1, 3, 4, 5, 128, 33, 24, 52]

    print(get_lost_cells_between_first_and_second_set(cells_at_t0, cells_at_t1))
    print(get_lost_cells_between_first_and_second_set(cells_at_t1, cells_at_t0))

    img = Img(
        '/E/Sample_images/segmentation_assistant/ovipo_uncropped/200709_armGFP_suz_46hAPF_ON.lif - Series011/tracked_cells_resized.tif')
    img = RGB_to_int24(img)
    start_all = timer()
    print(len(get_cells_in_image(img)))  # ça marche et c'est super rapide en fait --> really good and fast
    print('time', timer() - start_all)
    # can get the first pixel of each using indices

    start_all = timer()
    cells, first_px = get_cells_in_image_n_fisrt_pixel(
        img)  # ça marche et c'est super rapide en fait --> really good and fast
    print(cells)
    print(first_px)  # this is the ravel index
    print('time', timer() - start_all)

    # --> 0.20 secs
    start_all = timer()
    cells, area = get_cells_in_image_n_count(img)
    print(cells)
    print(area)
    print('time', timer() - start_all)

    # full self replacement of one image using == --> 14 secs
    start_all = timer()
    for cell in cells:
        img[img == cell] = cell
    print('time', timer() - start_all)

    # start_all = timer()
    # for cell in cells:
    #
    #     img[img == cell] = cell
    # print('time', timer() - start_all)

    # encore plus lent et null --> 45 secs
    # start_all = timer()
    # for cell in cells:
    #     np.where(img == cell, cell, img)
    #     # img[img == cell] = cell
    # print('time', timer() - start_all)

    # super slow --> do not use np.where in fact cause super slow
    start_all = timer()
    for cell in cells:
        cond = np.where(img == cell)
        img[cond] = cell
    print('time', timer() - start_all)

    # print('neighbs length', len(ids))

    # print(vertices)
    # if vertices[0].size != 0:
    #     print('y', vertices[0][0])
    # if vertices[1].size != 0:
    #     print('x', vertices[1][0])
    # if vertices[0].size + vertices[1].size != 0:
    #     print('first vx', result[vertices[0][0], vertices[1][0]])



# from https://www.learnpyqt.com/courses/concurrent-execution/multithreading-pyqt-applications-qthreadpool/
import os
from epyseg.settings.global_settings import set_UI
set_UI()
from qtpy.QtCore import Signal
from qtpy.QtWidgets import *
from qtpy.QtCore import *
import traceback
import sys


class WorkerSignals(QObject):
    """
    Defines the signals available from a running worker thread.

    Signals:
        finished: No data
        error: Tuple (exctype, value, traceback.format_exc())
        result: Any data returned from processing
        progress: Integer indicating % progress
    """

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class Worker(QRunnable):
    """
    Worker thread class.

    Inherits from QRunnable to handle worker thread setup, signals, and wrap-up.

    Args:
        fn (function): The function callback to run on this worker thread. Supplied args and kwargs will be passed through to the runner.
        args: Arguments to pass to the callback function
        kwargs: Keywords to pass to the callback function
    """

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (reused for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

        # Add the callback to our kwargs
        self.kwargs['progress_callback'] = self.signals.progress

    @Slot()
    def run(self):
        """
        Initialize the runner function with passed args and kwargs.
        """
        try:
            result = self.fn(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)  # Return the result of the processing
        finally:
            self.signals.finished.emit()  # Done




import gzip
import shutil
import os
import subprocess
from timeit import default_timer as timer
from epyseg.ta.tracking.tools import smart_name_parser

def compress_file_using_python(file_path, output_name=None):
    """
    Compresses the specified file using gzip compression.

    Args:
        file_path (str): The path to the file to be compressed.
        output_name (str, optional): The name of the compressed output file. If not provided, a default name is generated.
    """
    if output_name is None:
        output_name = os.path.splitext(file_path)[0] + '.gz'
    os.makedirs(smart_name_parser(output_name, 'parent'), exist_ok=True)
    with open(file_path, 'rb') as f_in:
        with gzip.open(output_name, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)


def compress_file_3rd_party(file_path, output_path=None):
    """
    Compresses the specified file using system gzip compression.

    Args:
        file_path (str): The path to the file to be compressed.
        output_path (str, optional): The path to the output file. If not provided, a default path is generated.
    """
    if output_path is None:
        output_path = file_path + '.gz'

    command = ['gzip', '-c', file_path]
    print('running: ' + ' '.join(command) + ' > ' + output_path)
    # create dirs if not exist
    os.makedirs(smart_name_parser(output_path, 'parent'), exist_ok=True)
    subprocess.run(['gzip', '-c', file_path], stdout=open(output_path, 'wb'))


if __name__ == '__main__':

    if True:
        # do a real compression of a real file
        start = timer()
        file_path = '/D/hiC_microC/hiC_data_fub1_n_ctrl/mapped.pairs'
        output_name = '/D/hiC_microC/hiC_data_fub1_n_ctrl/12-16.txt.gz'
        compress_file_3rd_party(file_path, output_name)
        print(timer() - start)
        import sys

        sys.exit(0)

    if False:
        # very slow --> only for small files
        start = timer()
        file_path = '/D/hiC_microC/hiC_data_fub1_n_ctrl/mapped.txt'
        compress_file_using_python(file_path)
        print(timer() - start)

        start = timer()
        output_name = '/D/hiC_microC/hiC_data_fub1_n_ctrl/tutu/mapped2.txt.gz'
        compress_file_using_python(file_path, output_name)
        print(timer() - start)

    if True:
        file_path = '/D/hiC_microC/hiC_data_fub1_n_ctrl/mapped.txt'
        output_path = '/D/hiC_microC/hiC_data_fub1_n_ctrl/tutu/mapped2.txt.gz'
        start = timer()
        compress_file_3rd_party(file_path, output_path)
        print(timer() - start)

        start = timer()
        compress_file_3rd_party(file_path)
        print(timer() - start)



import math
from skimage.measure import label, regionprops
from timeit import default_timer as timer
import numpy as np
import matplotlib.pyplot as plt

def add_2d_border(image2d):
    """
    Adds a border to a 2D image.

    Args:
        image2d (ndarray): The 2D image to which the border will be added.

    Returns:
        ndarray: The image with the added border.

    Examples:
        >>> image2d = np.array([[1, 2, 3],
        ...                     [4, 5, 6],
        ...                     [7, 8, 9]])
        >>> result = add_2d_border(image2d)
        >>> print(result)
        [[0 0 0]
         [0 5 0]
         [0 0 0]]
    """
    # Define the region where the border will be inserted
    insertHere = (slice(1, image2d.shape[0] - 1), slice(1, image2d.shape[1] - 1))

    # Create an array filled with zeros to hold the image with the border
    cells_with_borders = np.zeros_like(image2d)
    cells_with_borders.fill(0)

    # Copy the interior of the image to the array with the border
    cells_with_borders[insertHere] = image2d[insertHere]

    # Update the original image with the image containing the border
    image2d = cells_with_borders

    return image2d



def create_horizontal_gradient(cells):
    """
    Creates a horizontal gradient based on the cell IDs in a 2D array.

    Args:
        cells (ndarray): The 2D array containing cell IDs.

    Returns:
        ndarray: The horizontal gradient array.

    # Examples:
    #     >>> cells = np.array([[0, 0, 0],
    #     ...                   [1, 1, 0],
    #     ...                   [2, 2, 0]])
    #     >>> result = create_horizontal_gradient(cells)
    #     >>> print(result)
    #     [[0 0 0]
    #      [0 1 0]
    #      [0 0 0]]

    """
    # Add a border to the cells array
    cells = add_2d_border(cells)

    # Create an array of zeros with the same shape as cells to hold the horizontal gradient
    horiz_gradient = np.zeros_like(cells)

    # Iterate over the rows and columns of the cells array
    for j in range(0, cells.shape[-2], 1):
        counter = 1
        last_id = 0

        for i in range(0, cells.shape[-1], 1):
            # Check the current cell ID and compare it with the previous cell ID

            current_cell_id = cells[j][i]

            if current_cell_id != 0 and current_cell_id != last_id:
                last_id = current_cell_id

            if current_cell_id == 0 and last_id != 0:
                last_id = 0
                # Divide the counter by 2
                half_counter = int(counter / 2.)
                counter = 1  # Restart the counter

                # Found a new cell edge, go backwards to create the gradient
                reverse_counter = 1
                for ii in range(0, -half_counter + 1, -1):
                    horiz_gradient[j][i + ii] = reverse_counter
                    reverse_counter += 1

            if current_cell_id != 0 and current_cell_id == last_id:
                horiz_gradient[j][i] = counter
                counter += 1

    return horiz_gradient



def create_vertical_gradient(cells):
    """
    Creates a vertical gradient based on the cell IDs in a 2D array.

    Args:
        cells (ndarray): The 2D array containing cell IDs.

    Returns:
        ndarray: The vertical gradient array.

    # Examples:
    #     >>> cells = np.array([[0, 0, 0],
    #     ...                   [1, 1, 0],
    #     ...                   [2, 2, 0]])
    #     >>> result = create_vertical_gradient(cells)
    #     >>> print(result)
    #     [[0 0 0]
    #      [0 1 0]
    #      [0 0 0]]
    """
    # Add a border to the cells array
    cells = add_2d_border(cells)

    # Create an array of zeros with the same shape as cells to hold the vertical gradient
    vertical_gradient = np.zeros_like(cells)

    # Iterate over the columns and rows of the cells array
    for i in range(0, cells.shape[-1], 1):
        counter = 1
        last_id = 0

        for j in range(0, cells.shape[-2], 1):
            current_cell_id = cells[j][i]

            if current_cell_id != 0 and current_cell_id != last_id:
                last_id = current_cell_id

            if current_cell_id == 0 and last_id != 0:
                last_id = 0
                # Divide the counter by 2
                half_counter = int(counter / 2.)
                counter = 1  # Restart the counter

                # Found a new cell edge, go backwards to create the gradient
                reverse_counter = 1
                for jj in range(0, -half_counter + 1, -1):
                    vertical_gradient[j + jj][i] = reverse_counter
                    reverse_counter += 1

            if current_cell_id != 0 and current_cell_id == last_id:
                vertical_gradient[j][i] = counter
                counter += 1

    return vertical_gradient


def get_seeds(cells, one_seed_per_cell=True):
    """
    Obtain seeds for watershed segmentation based on cell gradients.

    Args:
        cells (ndarray): The 2D array containing cell IDs.
        one_seed_per_cell (bool): Whether to ensure only one seed per cell. Default is True.

    Returns:
        tuple: A tuple containing the combined gradients array and the seeds array.

    # Examples:
    #     >>> cells = np.array([[0, 0, 0],
    #     ...                   [1, 1, 0],
    #     ...                   [2, 2, 0]])
    #     >>> result = get_seeds(cells)
    #     >>> combined_gradients, seeds = result
    #     >>> print(combined_gradients)
    #     [[0 0 0]
    #      [1 2 0]
    #      [1 2 0]]
    #     >>> print(seeds)
    #     [[0 0 0]
    #      [0 2 0]
    #      [0 0 0]]
    """
    # Create horizontal and vertical gradients
    horiz_gradient = create_horizontal_gradient(cells)
    vertical_gradient = create_vertical_gradient(cells)

    # Perform further processing on gradients to obtain combined gradients and seeds
    combined_gradients_ready_for_wshed, seeds = _get_seeds(cells, horiz_gradient, vertical_gradient)

    # If one_seed_per_cell is True, ensure only one seed per cell
    if one_seed_per_cell:
        new_seeds = label(seeds.astype(np.uint8), connectivity=1, background=0)
        props_seeds = regionprops(new_seeds)

        extra_seeds_to_remove = []

        for region in regionprops(cells):
            cells_found = []
            for coordinates in region.coords:
                cells_found.append(new_seeds[coordinates[0], coordinates[1]])
            cells_found = list(dict.fromkeys(cells_found))
            if 0 in cells_found:
                cells_found.remove(0)
            if len(cells_found) > 1:
                max_area = 0
                for cell in cells_found:
                    region = props_seeds[cell - 1]
                    max_area = max(max_area, region.area)

                for cell in cells_found:
                    if not props_seeds[cell - 1].area == max_area:
                        extra_seeds_to_remove.append(cell)

        for region in props_seeds:
            if region.label in extra_seeds_to_remove:
                for coordinates in region.coords:
                    seeds[coordinates[0], coordinates[1]] = 0

    return combined_gradients_ready_for_wshed, seeds



def _get_seeds(cells, horiz_gradient, vertical_gradient):
    """
    Obtain seeds based on combined gradients for watershed segmentation.

    Args:
        cells (ndarray): The 2D array containing cell IDs.
        horiz_gradient (ndarray): The 2D array representing the horizontal gradient.
        vertical_gradient (ndarray): The 2D array representing the vertical gradient.

    Returns:
        tuple: A tuple containing the combined gradients array and the highest pixel values array.

    # Examples:
    #     >>> cells = np.array([[0, 0, 0],
    #     ...                   [1, 1, 0],
    #     ...                   [2, 2, 0]])
    #     >>> horiz_gradient = np.array([[0, 0, 0],
    #     ...                            [1, 2, 0],
    #     ...                            [1, 2, 0]])
    #     >>> vertical_gradient = np.array([[0, 0, 0],
    #     ...                              [0, 1, 0],
    #     ...                              [0, 1, 0]])
    #     >>> result = _get_seeds(cells, horiz_gradient, vertical_gradient)
    #     >>> combined_gradients, highest_pixels = result
    #     >>> print(combined_gradients)
    #     [[0 0 0]
    #      [1 3 0]
    #      [1 3 0]]
    #     >>> print(highest_pixels)
    #     [[  0   0   0]
    #      [  0 255   0]
    #      [  0 255   0]]
    """
    # Combine gradients
    combined_gradients_ready_for_wshed = horiz_gradient + vertical_gradient

    # Create array for highest pixel values
    highest_pixels = np.zeros_like(cells)

    # Iterate over labeled regions in cells with intensity_image set to combined_gradients_ready_for_wshed
    for region in regionprops(cells, intensity_image=combined_gradients_ready_for_wshed):
        max_val = 0

        # Scale seed by cell size using the square root of the region's area
        factor = math.sqrt(region.area) * 30 / 100

        # Find the maximum value in the intensity image
        for j in range(0, region.intensity_image.shape[-2], 1):
            for i in range(0, region.intensity_image.shape[-1], 1):
                if region.intensity_image[j, i] >= max_val:
                    max_val = region.intensity_image[j, i]

        # Set the region's image to False
        region.image.fill(False)

        # Set pixels in the region's image to True where the intensity exceeds (max_val - factor)
        region.image[region.intensity_image >= (max_val - factor)] = True

        # Update the highest_pixels array with the region's coordinates
        for coordinates in region.coords:
            highest_pixels[coordinates[0], coordinates[1]] = 255

    return combined_gradients_ready_for_wshed, highest_pixels



if __name__ == '__main__':
    start = timer()



from skimage.filters import threshold_sauvola
from skimage.morphology import remove_small_objects
import matplotlib.pyplot as plt
from epyseg.img import Img
import numpy as np
from epyseg.postprocess.superpixel_methods import get_optimized_mask2


def sauvola(img, window_size=25, min_threshold=0.02):
    """
    Applies the Sauvola thresholding method to an image.

    Args:
        img (numpy.ndarray): The input image.
        window_size (int): The size of the window for local thresholding. Defaults to 25.
        min_threshold (float): The minimum threshold value. Defaults to 0.02.

    Returns:
        numpy.ndarray: The thresholded image.

    """
    k = 0.25
    r = 0.5

    t = threshold_sauvola(img, window_size=window_size, k=k, r=r)
    if min_threshold is not None:
        t[t <= min_threshold] = min_threshold
    return t


def segment_cells(image, __DEBUG=False, __VISUAL_DEBUG=False, stop_at_threshold_step=False,
                  min_unconnected_object_size=None, min_threshold=None, window_size=25, real_avg_mode=False):
    """
    Segments cells in an image using the Sauvola thresholding method.

    Args:
        image (numpy.ndarray): The input image.
        __DEBUG (bool): Debug flag. Defaults to False.
        __VISUAL_DEBUG (bool): Visual debug flag. Defaults to False.
        stop_at_threshold_step (bool): Flag to stop at thresholding step and return the thresholded image. Defaults to False.
        min_unconnected_object_size (int): The minimum size of unconnected objects to be removed. Defaults to None.
        min_threshold (float): The minimum threshold value. Defaults to None.
        window_size (int): The size of the window for local thresholding. Defaults to 25.
        real_avg_mode (bool): Flag indicating whether to use real average mode. Defaults to False.

    Returns:
        numpy.ndarray: The segmented image.

    """
    original = image.copy()

    t = sauvola(image, min_threshold=min_threshold, window_size=window_size)

    if __VISUAL_DEBUG:
        plt.imshow(t)
        plt.show()

    image[image >= t] = 1
    image[image < t] = 0

    if __VISUAL_DEBUG and min_unconnected_object_size == 12:
        plt.imshow(image)
        plt.title('before')
        plt.show()

    if min_unconnected_object_size is not None and min_unconnected_object_size >= 1:
        image = remove_small_objects(image.astype(bool), min_size=min_unconnected_object_size, connectivity=2).astype(np.uint8)
        if __VISUAL_DEBUG and min_unconnected_object_size == 12:
            plt.imshow(image)
            plt.title('after')
            plt.show()

    if stop_at_threshold_step:
        image = image * 255
        return image

    return get_optimized_mask2(original, sauvola_mask=None, score_before_adding=True)


if __name__ == '__main__':
    from timeit import default_timer as timer

    # image = Img('/D/final_folder_scoring/predict_hybrid/mini_test.tif')
    # image = Img('/D/final_folder_scoring/predict_hybrid/AVG_StackFocused_Endocad-GFP(6-12-13)#19_000.tif')
    # image = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/122.tif')[...,0]
    # image = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/image_plant_best-zoomed.tif')[...,0]
    image = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/5.tif')[..., 0]
    # image = Img('/D/final_folder_scoring/predict_hybrid/tmp11.png')
    # image = Img('/D/final_folder_scoring/predict_hybrid/11-1_nuclei_1.tif')
    start = timer()

    # final_mask = segment_cells(image, __DEBUG=False, __VISUAL_DEBUG=False, stop_at_threshold_step=False)
    final_mask = segment_cells(image, __DEBUG=False, __VISUAL_DEBUG=False, stop_at_threshold_step=False, min_unconnected_object_size=12)

    duration = timer() - start
    print(duration)

    plt.imshow(final_mask)
    plt.show()
    Img(final_mask, dimensions='hw').save('/home/aigouy/Bureau/trash/test_new_seeds_seg_stuff/final_rewatershed.tif')



from epyseg.img import int24_to_RGB, RGB_to_int24
from timeit import default_timer as timer
import numpy as np
from epyseg.img import Img
# really very good --> now make it recursive then done --> extract the cell from there and continue
from natsort import natsorted
import os

# parameters
from epyseg.utils.loadlist import loadlist
from epyseg.ta.tracking.rapid_detection_of_vertices import associate_vertices_to_cells, detect_vertices_and_bonds

__DEBUG__ = False



def _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1, vertices_t0=None, vertices_t1=None, MIN_CUT_OFF_SWAPPING=None):
    """
    Fixes swapping of cells between two time steps.

    Args:
        tracked_cells_t0 (numpy.ndarray): Array of tracked cells at time t0.
        tracked_cells_t1 (numpy.ndarray): Array of tracked cells at time t1.
        vertices_t0 (numpy.ndarray, optional): Array of vertices at time t0. Defaults to None.
        vertices_t1 (numpy.ndarray, optional): Array of vertices at time t1. Defaults to None.
        MIN_CUT_OFF_SWAPPING (float, optional): Minimum cutoff score for swapping. Defaults to None.

    Returns:
        numpy.ndarray: Updated array of tracked cells at time t1.

    """
    if __DEBUG__:
        cell_to_check = 0x29d785  # 0x976494 #0x29d785  # 0x746502 #0x96c9ef #0x542358  # 0xf36619 #0x496bbd #0x1bf38f
        # map           9921684   2742149   7628034  9882095   5514072   15951385   4811709
        print('cell to check', cell_to_check, format(cell_to_check, 'x'))


    # inner_time = timer()
    if vertices_t0 is None:
        vertices_t0 = detect_vertices_and_bonds(tracked_cells_t0)
        vertices_t0 = np.where(vertices_t0 == 255)
    if vertices_t1 is None:
        vertices_t1 = detect_vertices_and_bonds(tracked_cells_t1)
        vertices_t1 = np.where(vertices_t1 == 255)
    # print('current time0', timer() - inner_time)

    # inner_time = timer()
    t0_cells_n_vertices = associate_vertices_to_cells(tracked_cells_t0, vertices_t0,
                                                      forbidden_colors=(0, 0xFFFFFF),
                                                      output_cells_and_their_neighbors=True)
    t1_cells_n_vertices = associate_vertices_to_cells(tracked_cells_t1, vertices_t1,
                                                      forbidden_colors=(0, 0xFFFFFF),
                                                      output_cells_and_their_neighbors=True)


    if __DEBUG__:
        print('t0_cells_n_vertices', t0_cells_n_vertices)
        print('t1_cells_n_vertices', t1_cells_n_vertices)

    # inner_time = timer()
    best_pairs = {}  # dict that contains pairs -->  a good idea
    best_pairs_score = {}

    for key, value in t0_cells_n_vertices.items():
        max_match = 0.
        best_match = 0
        neighs_at_t0 = t0_cells_n_vertices[key]
        best_pair_size = 0

        for key2, value2 in t1_cells_n_vertices.items():
            neighs_at_t1 = t1_cells_n_vertices[key2]
            common_neighs = set(neighs_at_t1).intersection(neighs_at_t0)

            # if two are equal --> try to take the best --> find a way to do it smartly
            if len(common_neighs) > 0 and len(
                    common_neighs) >= max_match:  # if found twice --> find a way to select which cells to take
                # max_match = len(common_neighs) / ((len(neighs_at_t1)+len(neighs_at_t0))/2)

                conti = True
                if len(common_neighs) == max_match:
                    if len(neighs_at_t1) == best_pair_size:
                        if key == key2:
                            # do favor identity as much as possible
                            conti = True
                        # find a trick or random
                        elif key2 in neighs_at_t0:  # not sure this is very smart but give it a try NB recursion may fix bugs
                            conti = False
                    elif len(neighs_at_t1) > best_pair_size:
                        conti = False
                    # si les 2 font la meme taille --> pb à nouveau --> y a t'il une astuce...

                if conti:
                    max_match = len(common_neighs)  # / ((len(neighs_at_t1)+len(neighs_at_t0))/2)
                    best_match = key2
                    best_pair_size = len(neighs_at_t1)
                    if __DEBUG__:
                        if key == cell_to_check:
                            hex_list_0 = [format(x, 'x') for x in neighs_at_t0]
                            hex_list_1 = [format(x, 'x') for x in neighs_at_t1]
                            print('pairs for a test', neighs_at_t0, neighs_at_t1, len(common_neighs), max_match,
                                  format(key, 'x'), format(key2, 'x'), key, key2, hex_list_0,
                                  hex_list_1)  # that works --> so indeed the best match is found but then cells are lost and swapped --> why is that --> I still miss a ctrl somewhere --> or a secondary swap that is less good somehow --> check that or a coding error!!
                else:
                    if __DEBUG__:
                        if key == cell_to_check:
                            hex_list_0 = [format(x, 'x') for x in neighs_at_t0]
                            hex_list_1 = [format(x, 'x') for x in neighs_at_t1]
                            print('discarded pairs for a test', neighs_at_t0, neighs_at_t1, len(common_neighs),
                                  max_match,
                                  format(key, 'x'), format(key2, 'x'), key, key2, hex_list_0,
                                  hex_list_1)
        if __DEBUG__:
            if key == cell_to_check:
                print('final match', max_match, cell_to_check, best_match, format(cell_to_check, 'x'),
                      format(best_match, 'x'))
        # or best --> compare to its current match and if better then take it otherwise ignore  --> maybe

        # in fact I take all here but is that really worth ??? --> maybe take one but only if the cell has just one neighbor ???? --> shall I start with 2 ??? that would dramatically reduce the nb of cells I guess --> think about it
        if max_match >= 0.5:  # take all cells that have at least half of their neighbs --> 40--> maybe ok

            # compare neighbors of current match versus those in the previous
            best_pairs[key] = best_match
            best_pairs_score[key] = max_match  # will be used later to avoid issues

    # remove cells that did not change much
    if MIN_CUT_OFF_SWAPPING is not None and MIN_CUT_OFF_SWAPPING>0:
        pairs_to_ignore = [k for k, v in best_pairs_score.items() if v <= MIN_CUT_OFF_SWAPPING]
        for ids in pairs_to_ignore:
            del best_pairs[ids]

    # finding duplicate values
    # from dictionary using set
    rev_dict = {}
    for key, value in best_pairs.items():
        rev_dict.setdefault(value, set()).add(key)

    dupes_matches_to_be_removed = []
    for key, value in rev_dict.items():

        if __DEBUG__:
            if key == cell_to_check:
                print('MEGA THERE######')

        if len(value) > 1:
            best_match = 0
            max_match = 0
            # keep only best and remove all others from the table
            for iii, val in enumerate(value):
                if __DEBUG__:
                    if key == cell_to_check:
                        print(iii, val, best_pairs_score[val])

                if best_pairs_score[val] == max_match:
                    if val == key:
                        max_match = best_pairs_score[val]
                        best_match = val
                elif best_pairs_score[val] > max_match:
                    max_match = best_pairs_score[val]
                    best_match = val

            if __DEBUG__:
                if key == cell_to_check:
                    print(value, val, max_match)

            value.remove(best_match)
            # I do have a bug here but I don't get it in fact but ok for now
            dupes_matches_to_be_removed.extend(
                value)  # not a good idea --> just need to remove it from the value not from everywhere... is that correct ???? remove it only if present again in duplicates in final stuff after cleaning

    if __DEBUG__:
        print('dupes_matches_to_be_removed', dupes_matches_to_be_removed,
              [format(x, 'x') for x in dupes_matches_to_be_removed])

    for key in dupes_matches_to_be_removed:
        if key in best_pairs:
            del best_pairs[key]

    def swap(pairs_of_cells_to_swap,
             tracked_cells_t1):  # this is super slow and would benefit from recoding it in C --> this takes most of the time of the function
        # inner_time = timer()
        clone_tracks = np.array(tracked_cells_t1, copy=True)
        # print('current time', timer() - inner_time)

        for cell1, cell2 in pairs_of_cells_to_swap.items():
            if cell1 == cell2: # remove identity cause it slows down a lot the process
                continue

            # if cell2 not in pairs_of_cells_to_swap.keys(): # very slow --> find a way to make that faster
            clone_tracks[tracked_cells_t1 == cell1] = cell2
            clone_tracks[tracked_cells_t1 == cell2] = cell1

            if __DEBUG__:
                print('converting', cell2, '--> ', cell1, format(cell2, 'x'), '-->', format(cell1, 'x'))

            # find a way todo things and to not duplicate stuff

            # pb --> this creates duplications but otherwise it does swap correct cells --> find a way to prevent this --> I need to order swaps and other stuff
        # print('current time2', timer() - inner_time)
        return clone_tracks

    # inner_time = timer()
    tracked_cells_t1 = swap(best_pairs,
                            tracked_cells_t1)  # c'est ce truc qui est super lent --> plus de la moitié du temps --> try improve that

    if __DEBUG__:
        print('best_pairs', len(best_pairs), best_pairs)  # maybe best ratio versus its own size
    # TODO maybe save swapped cells --> only show those and blacken all the rest

    return tracked_cells_t1


if __name__ == '__main__':

    swap_fix_recursions = 3

    # pb is because after translation the cell is also present in both --> skipped by my stuff --> need detect those cells and chose to get only one in fact
    # pb is because after translation the cell is also present in both --> skipped by my stuff --> need detect those cells and chose to get only one in fact
    # or first do the

    start_all = timer()

    # path = '/E/Sample_images/tracking_test/leg_uncropped'  # quick test of the tracking algo
    # path = '/E/Sample_images/tracking_test/leg_cropped'  # quick test of the tracking algo
    path = '/E/Sample_images/tracking_test/test_uncropped'  # quick test of the tracking algo
    # path = '/E/Sample_images/tracking_test/test_cropped'  # quick test of the tracking algo

    if not path.lower().endswith('.txt') and not path.lower().endswith('.lst'):
        images_to_analyze = os.listdir(path)
        images_to_analyze = [os.path.join(path, f) for f in images_to_analyze if
                             os.path.isfile(os.path.join(path, f))]  # list only files and only if they end by tif
        images_to_analyze = natsorted(images_to_analyze)
    else:
        images_to_analyze = loadlist(path)

    # get current and next in fact and map next

    # process first image directly then move on to next
    # images_to_analyze = []

    # for image 0 --> just randomly fill it with colors --> make sure no dupe and that's it

    # TODO --> do the first

    print(images_to_analyze)

    for l in range(len(images_to_analyze) - 1):

        start_loop = timer()

        file_path_0 = images_to_analyze[l]
        file_path_1 = images_to_analyze[l + 1]

        print('files', file_path_1, file_path_0)

        if not file_path_1.endswith('.tif') or not file_path_0.endswith('.tif'):
            continue

        # print('files', file_path_1, file_path_0)

        filename0_without_ext = os.path.splitext(file_path_0)[0]
        filename1_without_ext = os.path.splitext(file_path_1)[0]

        tracked_cells_t0 = Img(os.path.join(filename0_without_ext, 'tracked_cells_resized.tif'))
        tracked_cells_t1 = Img(os.path.join(filename1_without_ext, 'tracked_cells_resized.tif'))

    vertices_t0 = detect_vertices_and_bonds(RGB_to_int24(tracked_cells_t0))
    vertices_t0 = np.where(vertices_t0==255)
    vertices_t1 = detect_vertices_and_bonds(RGB_to_int24(tracked_cells_t1))
    vertices_t1 = np.where(vertices_t1 == 255)

    # time is always increasing but why is that is some set not reset
    for fix in range(swap_fix_recursions):
        print('swapping fix recursion #', fix)
        begin_loop = timer()
        tracked_cells_t1 = _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1, vertices_t0=vertices_t0, vertices_t1=vertices_t1)
        # tracked_cells_t1 = _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1)
        print('current time', timer() - begin_loop)
        # Img(tracked_cells_t1).save(os.path.join(filename0_without_ext, 'track_after_swap_correction'+ str(fix)+ '.tif'), mode='raw')

    # example of swapped cells
    print('total time', timer() - start_all)  # --> 47 secs --> un peu lent qd meme

    Img(int24_to_RGB(tracked_cells_t1)).save(os.path.join(filename0_without_ext, 'track_after_swap_correction.tif'), mode='raw')

    # puis je la faire dynamiquement pr eviter les swaps en live
    # --> void aussi si rentable ou pas

    # TODO -->
    # this now seems ok and could be done recursively...
    # could also use neighbor cells to reassign most likely cell --> cell that matches most of the surrounding IDs

    # pas mal --> see how many recursions might be needed to get the stuff to work --> might be time consuming to do plenty of recursion... --> maybe just update the map so that not everything needs be calculated again

    # if a cell is unsure try reassign it by its neighborood...

    # TODO --> reidentify cells by neighborhood --> simply use the vertices to detect the neighbors --> split the code



import traceback
import matplotlib.pyplot as plt

from epyseg.ta.measurements.measurements3D.get_point_on_surface_if_centroid_is_bad import point_on_surface
from epyseg.ta.utils.TA_tools import get_TA_clone_file_name
from epyseg.img import RGB_to_int24, Img
from epyseg.ta.tracking.tools import smart_name_parser
from skimage.measure import label, regionprops
import numpy as np
from epyseg.tools.early_stopper_class import early_stop
from epyseg.tools.logger import TA_logger  # logging

logger = TA_logger()

# deprecated --> use another code if possible
def convert_db_xy_coordiantes_to_local_cell_ID_for_clones(db_path, table_name, image_for_correspondance_label_or_RGB24, return_image=False):
    """
    Converts database coordinates to local cell IDs for clones.

    Args:
        db_path (str): The path to the database.
        table_name (str): The name of the table containing the coordinates.
        image_for_correspondance_label_or_RGB24 (ndarray): The image used for correspondence between labels and RGB24 values.
        return_image (bool, optional): A flag indicating whether to return an image or a set of cells.
                                       Defaults to False.

    Returns:
        list or ndarray: The output, either a list of cells or an image.

    """
    from epyseg.ta.database.sql import TAsql
    import numpy as np

    output = []  # Initialize the output variable
    db = None
    try:
        db = TAsql(db_path)  # Open the database connection

        # Get the coordinates from the database
        query_results = db.run_SQL_command_and_get_results('SELECT * FROM ' + table_name, return_header=False)

        if return_image:
            output = np.zeros_like(image_for_correspondance_label_or_RGB24)  # Create an empty image

        for coords in query_results:
            color = image_for_correspondance_label_or_RGB24[coords[0], coords[1]]  # Get the color value at the coordinates

            if return_image:
                output[image_for_correspondance_label_or_RGB24 == color] = color  # Assign the color to corresponding pixels in the image
            else:
                output.append(color)  # Append the color to the list of cells
    except:
        import traceback
        traceback.print_exc()
        logger.error('Something went wrong, selection could not be converted to 2D coordinates')
    finally:
        if db is not None:
            try:
                db.close()  # Close the database connection
            except:
                import traceback
                traceback.print_exc()

    return output

def get_colors_drawn_over(mask, image_to_analyze, forbidden_colors=[0, 0xFFFFFF]):
    """
    Retrieves the colors drawn over a mask in an image.

    Args:
        mask (ndarray): The mask indicating the drawn areas.
        image_to_analyze (ndarray): The image to analyze.
        forbidden_colors (list or int, optional): Colors to exclude from the result.
                                                  Defaults to [0, 0xFFFFFF].

    Returns:
        list: The list of colors drawn over the mask.

    """
    if mask is None:
        return None

    selected_colors = image_to_analyze[mask != 0]  # Get the colors corresponding to the non-zero regions of the mask

    if len(image_to_analyze.shape) == 3:
        selected_colors = RGB_to_int24(selected_colors)  # Convert RGB colors to int24 format

    selected_colors = set(selected_colors.ravel().tolist())  # Convert the colors to a set

    if forbidden_colors is not None:
        if not isinstance(forbidden_colors, list):
            forbidden_colors = [forbidden_colors]  # Convert single forbidden color to a list

        for color in forbidden_colors:
            if color in selected_colors:
                selected_colors.remove(color)  # Remove forbidden colors from the selected colors set

    selected_colors = list(selected_colors)  # Convert the set back to a list

    return selected_colors


def handCorrection_to_label(handCorrection):
    """
    Converts a hand correction image to a labeled image.

    Args:
        handCorrection (str or ndarray): The hand correction image as a filename or ndarray.

    Returns:
        ndarray: The labeled image.

    """
    if isinstance(handCorrection, str):
        handCorrection = Img(handCorrection)  # Load the hand correction image if it's a filename

    if len(handCorrection.shape) == 3:
        handCorrection = handCorrection[..., 0]  # Convert RGB image to grayscale

    handCorrection = label(handCorrection, connectivity=1, background=255)  # Label the image

    return handCorrection

def add_coords_to_db(db_path, table_name_to_store_coords_to, coords):
    """
    Adds coordinates to a database table.

    Args:
        db_path (str): The path to the database.
        table_name_to_store_coords_to (str): The name of the table to store the coordinates.
        coords (list): A list of coordinate tuples in the format (y, x).

    """
    from epyseg.ta.database.sql import TAsql

    coords_of_selection = {'pixel_within_cell_y': [], 'pixel_within_cell_x': []}

    db = None
    try:
        db = TAsql(db_path)  # Open the database connection

        # Extract y and x coordinates from the coordinate tuples and store them in separate lists
        for coord in coords:
            coords_of_selection['pixel_within_cell_y'].append(coord[0])
            coords_of_selection['pixel_within_cell_x'].append(coord[1])

        # Create a new table or append to an existing table in the database with the coordinate data
        db.create_and_append_table(table_name=table_name_to_store_coords_to, datas=coords_of_selection)
    except:
        traceback.print_exc()
        logger.error('Something went wrong while trying to write coordinates to the database')
    finally:
        if db is not None:
            try:
                db.close()  # Close the database connection
            except:
                traceback.print_exc()


def convert_selection_color_to_coords(img_to_analyze_RGB24_or_label, selected_cells, bg_color=None):
    """
    Converts selected cell colors to coordinates within the image.

    Args:
        img_to_analyze_RGB24_or_label (numpy.ndarray): The image to analyze, either in RGB24 format or label format.
        selected_cells (list): A list of selected cell colors.
        bg_color (int, optional): The background color. If not provided, it will be determined automatically.

    Returns:
        list: A list of coordinate tuples representing the selected cells.

    """
    # tmp = img_to_analyze
    coords_of_selection = []

    if img_to_analyze_RGB24_or_label is None:
        logger.error('No image provided. Nothing to do...')
        return coords_of_selection

    if len(img_to_analyze_RGB24_or_label.shape) == 3:
        img_to_analyze_RGB24_or_label = RGB_to_int24(img_to_analyze_RGB24_or_label)

    if bg_color is None:
        if 0xFFFFFF in img_to_analyze_RGB24_or_label:
            bg_color = 0xFFFFFF
        else:
            bg_color = 0

    cell_label = label(img_to_analyze_RGB24_or_label, connectivity=1, background=bg_color)

    # For all the selected cells, return the coordinates
    for region in regionprops(cell_label):
        color = img_to_analyze_RGB24_or_label[region.coords[0][0], region.coords[0][1]]
        if color in selected_cells:
            point_always_inside_the_cell = point_on_surface(region, cell_label)
            coords_of_selection.append(point_always_inside_the_cell)

    return coords_of_selection


def convert_coords_to_IDs(img_to_analyze_RGB24_or_label, selected_coords, forbidden_colors=[0xFFFFFF, 0], return_image=False, new_color_to_give_to_cells_if_return_image=None):
    """
    Converts selected coordinates to cell IDs in the image.

    Args:
        img_to_analyze_RGB24_or_label (numpy.ndarray): The image to analyze, either in RGB24 format or label format.
        selected_coords (list): A list of selected coordinates.
        forbidden_colors (list, optional): Colors to exclude from conversion. Defaults to [0xFFFFFF, 0].
        return_image (bool, optional): Whether to return an image with converted cell IDs. Defaults to False.
        new_color_to_give_to_cells_if_return_image (int, optional): The color to assign to cells in the returned image. Only applicable if return_image is True.

    Returns:
        list or numpy.ndarray: A list of cell IDs or an image with converted cell IDs, depending on the return_image parameter.

    """
    ids = []

    if img_to_analyze_RGB24_or_label is None:
        logger.error('Empty image. Nothing to do...')
        if return_image:
            return None
        return ids

    if len(img_to_analyze_RGB24_or_label.shape) == 3:
        img_to_analyze_RGB24_or_label = RGB_to_int24(img_to_analyze_RGB24_or_label)

    if return_image:
        ids = np.zeros_like(img_to_analyze_RGB24_or_label)

    for coords in selected_coords:
        color = img_to_analyze_RGB24_or_label[coords[0], coords[1]]

        if forbidden_colors is not None:
            if color in forbidden_colors:
                continue

        if not return_image:
            ids.append(color)
        else:
            # Can be used to plot clones
            if new_color_to_give_to_cells_if_return_image is not None:
                ids[img_to_analyze_RGB24_or_label == color] = new_color_to_give_to_cells_if_return_image
            else:
                ids[img_to_analyze_RGB24_or_label == color] = color

    return ids

def convert_selected_cells_to_local_db_coords(selected_cells, label_or_RGB24_image_corresponding_to_selected_cells,
                                              db_path, table_name_to_store_coords_to, bg_color_for_label=None):
    """
    Converts selected cells to local coordinates and stores them in a database.

    Args:
        selected_cells (list): List of selected cell IDs.
        label_or_RGB24_image_corresponding_to_selected_cells (numpy.ndarray): The label or RGB24 image corresponding to the selected cells.
        db_path (str): Path to the database.
        table_name_to_store_coords_to (str): Name of the table to store the coordinates.
        bg_color_for_label (int, optional): Background color for label image. Defaults to None.

    Returns:
        None

    """
    if not selected_cells:
        logger.error('Nothing selected. Nothing to do.')
        return

    if bg_color_for_label is None:
        # Auto mode to determine the label
        if 0xFFFFFF in label_or_RGB24_image_corresponding_to_selected_cells:
            bg_color_for_label = 0xFFFFFF
        else:
            bg_color_for_label = 0

    try:
        cell_label = label(label_or_RGB24_image_corresponding_to_selected_cells, connectivity=1,
                           background=bg_color_for_label)

        coords = []
        for region in regionprops(cell_label):
            color = label_or_RGB24_image_corresponding_to_selected_cells[region.coords[0][0], region.coords[0][1]]
            if color in selected_cells:
                # Get the coordinate always inside the cell and store it in the database
                point_awlays_inside_the_cell = point_on_surface(region, cell_label)
                coords.append(point_awlays_inside_the_cell)

        # Store the coordinates in the database
        add_coords_to_db(db_path, table_name_to_store_coords_to, coords)

    except:
        traceback.print_exc()
        logger.error('Something went wrong. Selection could not be converted to 2D coordinates')


if __name__ == '__main__':
    import sys

    if False:
        # test how to append selection to a clone database
        path_to_image = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_empty/focused_Series012.png'
        selected_cells = [200, 201, 302]  # cells 1-3 have been selected
        handCorrection = Img(smart_name_parser(path_to_image, ordered_output='handCorrection.tif'))
        if len(handCorrection.shape) == 3:
            handCorrection = handCorrection[..., 0]
        cell_label = label(handCorrection, connectivity=1, background=255)
        # print(smart_name_parser(path_to_image, ordered_output='pyTA.db'))
        convert_selected_cells_to_local_db_coords(selected_cells, cell_label,
                                                  smart_name_parser(path_to_image, ordered_output='pyTA.db'),
                                                  'test_db_name', bg_color_for_label=0)  # smart
        # try relaod the coords as an image
        sys.exit(0)

    # ça marche vraiment bien en fait j'adore!!!
    if True:
        # load database from image and get the corresponding cell image or db depending on what is desired --> TODO
        path_to_image = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_empty/focused_Series012.png'
        handCorrection = Img(smart_name_parser(path_to_image, ordered_output='handCorrection.tif'))
        if len(handCorrection.shape) == 3:
            handCorrection = handCorrection[..., 0]
        cell_label = label(handCorrection, connectivity=1, background=255)
        output = convert_db_xy_coordiantes_to_local_cell_ID_for_clones(smart_name_parser(path_to_image, ordered_output='pyTA.db'), 'test_db_name', cell_label, return_image=True) #return_image=Fals
        # ça marche vraiment super bien
        if isinstance(output, np.ndarray):
            plt.imshow(output)
            plt.show()
        else:
            print(output)
        # --> this does return cells 1 to 3
        sys.exit(0)



import os
from epyseg.settings.global_settings import set_UI  # set the UI to be used py qtpy
set_UI()
import sys
from qtpy.QtWidgets import QApplication, QFileDialog
from os.path import expanduser


def openFileNameDialog(parent_window=None, extensions="Supported Files (*.jpg *.tif *.png);;All Files (*)",
                       path=expanduser('~')):
    """
    Opens a file dialog to select a single file.

    Args:
        parent_window (object): Parent window object. Default is None.
        extensions (str): File extension filters for the dialog. Default is "Supported Files (*.jpg *.tif *.png);;All Files (*)".
        path (str): Initial path for the dialog. Default is user's home directory.

    Returns:
        str: Selected file name.

    # Examples:
    #     >>> filename = openFileNameDialog()
    """
    if not os.path.exists(path):
        path = expanduser('~')
    fileName, _ = QFileDialog.getOpenFileName(parent_window, "Select a File", path, extensions,
                                              options=QFileDialog.DontUseNativeDialog)
    return fileName


def openFileNamesDialog(parent_window=None, extensions="Supported Files (*.jpg *.tif *.png);;All Files (*)",
                        path=expanduser('~')):
    """
    Opens a file dialog to select multiple files.

    Args:
        parent_window (object): Parent window object. Default is None.
        extensions (str): File extension filters for the dialog. Default is "Supported Files (*.jpg *.tif *.png);;All Files (*)".
        path (str): Initial path for the dialog. Default is user's home directory.

    Returns:
        list: List of selected file names.

    # Examples:
    #     >>> files = openFileNamesDialog()
    """
    if not os.path.exists(path):
        path = expanduser('~')
    files, _ = QFileDialog.getOpenFileNames(parent_window, "Select Files", path, extensions,
                                            options=QFileDialog.DontUseNativeDialog)
    return files


def saveFileDialog(parent_window=None, path=expanduser('~'), extensions="All Files (*);;Text Files (*.txt)",
                    default_ext=None):
    """
    Opens a file dialog to save a file.

    Args:
        parent_window (object): Parent window object. Default is None.
        path (str): Initial path for the dialog. Default is user's home directory.
        extensions (str): File extension filters for the dialog. Default is "All Files (*);;Text Files (*.txt)".
        default_ext (str): Default file extension. Default is None.

    Returns:
        str: Selected file name to save.

    # Examples:
    #     >>> filename = saveFileDialog(default_ext='.tif')
    """
    try:
        if not os.path.exists(path) and not os.path.exists(os.path.dirname(path)):
            path = expanduser('~')
    except:
        path = expanduser('~')
    fd = QFileDialog(parent_window, "Save a File", path, extensions, options=QFileDialog.DontUseNativeDialog)
    if default_ext is not None:
        fd.setDefaultSuffix(default_ext)
    fd.setAcceptMode(QFileDialog.AcceptSave)
    selected = fd.exec()
    if selected:
        fileName = fd.selectedFiles()[0]
        return fileName
    else:
        return


def openDirectoryDialog(parent_window=None, path=expanduser('~')):
    """
    Opens a dialog to select a directory.

    Args:
        parent_window (object): Parent window object. Default is None.
        path (str): Initial path for the dialog. Default is user's home directory.

    Returns:
        str: Selected directory path.

    # Examples:
    #     >>> folder = openDirectoryDialog()
    """
    if not os.path.exists(path):
        path = expanduser('~')
    folderName = QFileDialog.getExistingDirectory(parent_window, "Select a Directory", path,
                                                  options=QFileDialog.DontUseNativeDialog)
    if folderName is not None:
        if not folderName.strip():
            return None
        if not folderName.endswith("/") and not folderName.endswith("\\"):
            folderName += '/'
    return folderName


if __name__ == '__main__':
    app = QApplication(sys.argv)
    filename = saveFileDialog(default_ext='.tif')
    print(filename)
    sys.exit(0)
